module ZK
open List
open ST

(* A variant of signatures that doesn't transfer predicates *)

assume type data

type vk = data
type sk = data
type tag = data

assume val keygen : sk
assume val sk2vk  : sk -> Tot vk
assume val sign   : sk -> data -> tag
assume val verify : k:vk -> d:data -> t:tag -> Tot bool

(* Zero-knowledge proofs for chains of size 2 *)

(* Zero knowledge proof is represented as ref unit *)
type zkp = ref unit 
(* type tzk = (vk * vk * tag * data * tag) *)
type tzk = (vk * list vk * list tag * data * tag)

(* Reference for sealing *)
let zk_ref = alloc #(list (tzk*zkp)) []

(* environment to encode that we trust a set of keys *)

val extend : list vk -> vk -> Tot (list vk)
let extend g vk_x = append g [vk_x]

val is_trusted: l:list vk -> vkin:vk -> Tot (b:bool{b <==> mem vkin l})
let is_trusted g vk_maybe_trusted = mem vk_maybe_trusted g

(* Creating a zero knowledge proof using sealing *)
val mk_zk_2 : vk0:vk -> vk1:vk -> t1:tag -> m:data -> tm:tag -> zkp
let mk_zk_2 vk0 vk1 t1 m tm =
  let tuple = (vk0, [vk1], [t1], m, tm) in
  match assoc tuple (read zk_ref) with
  | Some zk_proof-> zk_proof
  | None ->
     let fresh = alloc unit () in
     write zk_ref ((tuple,fresh)::(read zk_ref));
     fresh

val swap : ('a*'b) -> ('b*'a)
let swap (a,b) = (b,a)

(* Verifying a zero knowledge proof using unsealing *)
val ver_zk_2 : vk0:vk -> m:data -> zkp ->
                      b:bool{b ==> (exists (vk1:vk) (t1:tag) (tm:tag).
                                          verify vk0 vk1 t1 /\ verify vk1 m tm)}
let ver_zk_2 vk0 m zk_proof =
  match assoc zk_proof (map swap (read zk_ref)) with
  | Some (vk0', [vk1], [t1], m', tm) -> vk0 = vk0' && m = m' &&
                                    verify vk0 vk1 t1 && verify vk1 m tm
  | Some _ -> false (* wrong format, not only one vk or not only one tag *)                                  
  | None -> false



val mk_zk_n : vk0:vk -> vks:list vk -> ts:list tag -> m:data -> tm:tag -> zkp
let mk_zk_n vk0 vks ts m tm =
  let tuple = (vk0, vks, ts, m, tm) in
  match assoc tuple (read zk_ref) with
  | Some zk_proof-> zk_proof
  | None ->
     let fresh = alloc unit () in
     write zk_ref ((tuple,fresh)::(read zk_ref));
     fresh

(* verifies the chain *)
val verify_n: vk0:vk -> vks:list vk -> ts:list tag -> m:data -> tm:tag -> Tot bool
 (decreases (length vks + length ts) )
let rec verify_n vk0 vks ts m tm = match (vks, ts) with
  | [], [] -> verify vk0 m tm
  | vki::vktl, tsi::tstl -> verify vk0 vki tsi && verify_n vki vktl tstl m tm
  | _ -> false (* vks and ts don't have equal length *)


val ver_zk : vk0:vk -> m:data -> zkp ->
                    b:bool{b ==> (exists (vks:list vk) (ts:list tag) (tm:tag).
                                           verify_n vk0 vks ts m tm)}
let ver_zk vk0 m zk_proof = 
match assoc zk_proof (map swap (read zk_ref)) with
  | Some (vk0', vks, ts, m', tm) -> vk0 = vk0' && m = m' &&
                                    verify_n vk0 vks ts m tm
  | None -> false

